"""Stateful programmatic WWW navigation, after Perl's WWW::Mechanize.

Copyright 2003 John J. Lee <jjl@pobox.com>
Copyright 2003 Andy Lester (original Perl code)

This code is free software; you can redistribute it and/or modify it under
the terms of the BSD License (see the file COPYING included with the
distribution).

"""

import urlparse, re

import ClientCookie
import ClientCookie._Util
import ClientForm
import pullparser
# serves me right for not using a version tuple...
VERSION_RE = re.compile(r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<bugfix>\d+)"
                        r"(?P<state>[ab])?(?:-pre)?(?P<pre>\d+)?$")
def parse_version(text):
    m = VERSION_RE.match(text)
    if m is None:
        raise ValueError
    return tuple([m.groupdict()[part] for part in
                  ("major", "minor", "bugfix", "state", "pre")])
assert map(int, parse_version(ClientCookie.VERSION)[:3]) >= [0, 4, 12], \
       "ClientCookie 0.4.12 or newer is required"
assert map(int, parse_version(ClientForm.VERSION)[:2]) >= [0, 1], \
       "ClientForm 0.1.x is required"
assert pullparser.__version__[:3] >= (0, 0, 2), \
       "pullparser 0.0.2a or newer is required"
del VERSION_RE, parse_version

from _useragent import UserAgent

__version__ = (0, 0, 2, "a", None)  # 0.0.2a

class BrowserStateError(Exception): pass
class LinkNotFoundError(Exception): pass
class FormNotFoundError(Exception): pass

class Link:
    def __init__(self, base_url, url, text, tag, attrs):
        assert None not in [url, tag, attrs]
        self.base_url = base_url
        self.absolute_url = urlparse.urljoin(base_url, url)
        self.url, self.text, self.tag, self.attrs = url, text, tag, attrs
    def __eq__(self, other):
        for name in "url", "text", "tag", "attrs":
            if getattr(self, name) != getattr(other, name):
                return False
        return True
    def __repr__(self):
        return "Link(base_url=%r, url=%r, text=%r, tag=%r, attrs=%r)" % (
            self.base_url, self.url, self.text, self.tag, self.attrs)

# XXXX when HTTPCookieProcessor is switched off in UserAgent, this
#  class stops working!
class Browser(UserAgent):
    """Browser-like class with support for history, forms and links.

    BrowserStateError is raised whenever the browser is in the wrong state to
    complete the requested operation - eg., when .back() is called when the
    browser history is empty, or when .follow_link() is called when the current
    response does not contain HTML data.

    Public attributes:

    request: last request (ClientCookie.Request or urllib2.Request)
    response: last response (as return value of urllib2.urlopen())
    form: currently selected form (see .select_form())

    """
    urltags = {
        "a": "href",
        "area": "href",
        "frame": "src",
        "iframe": "src",
    }

    def __init__(self):
        self._history = []  # LIFO
        self.request = self.response = None
        self.form = None
        self._forms = None
        self._title = None
        self._links = None
        UserAgent.__init__(self)  # do this last to avoid __getattr__ problems

    def open(self, url, data=None):
        if self.request is not None:
            self._history.append((self.request, self.response))

        try:
            url.get_full_url
        except AttributeError:
            # string URL -- convert to absolute URL if required
            scheme, netloc = urlparse.urlparse(url)[:2]
            if not scheme:
                # relative URL
                assert not netloc, "malformed URL"
                if self.response is None:
                    raise BrowserStateError(
                        "attempt to fetch relative URL before any URL has "
                        "been .open()ed")
                url = urlparse.urljoin(self.response.geturl(), url)
        # we want self.request to be assigned even if OpenerDirector.open fails
        self.request = self._request(url, data)

        self.response = ClientCookie.OpenerDirector.open(
            self, self.request, data)
        if not hasattr(self.response, "seek"):
            self.response = ClientCookie._Util.seek_wrapper(self.response)
        self._parse_html(self.response)

        return self.response

    def geturl(self): return self.response.geturl()

##     def open(self, url, data=None):
##         if self.request is not None:
##             self._history.append((self.request, self.response))
##         #print ".open()ing url", url
##         # we have to absolutize both string urls and Requests
##         self.request = self._request(url, data)
##         url = self.request.get_full_url()
##         scheme, netloc, path, parameters, query, frag = urlparse.urlparse(url)
##         if not scheme:
##             # relative URL
##             assert not netloc
##             if self.response is None:
##                 raise BrowserStateError("attempt to fetch relative URL before "
##                                         "any URL has been .open()ed")
##             url = urlparse.urljoin(self.response.geturl(), url)
##             self.request = self._request(url, data)
##         self.response = ClientCookie.OpenerDirector.open(
##             self, self.request, data)
##         self._parse_html(self.response)
##         return self.response

    def reload(self):
        """Reload current document, and return response object."""
        if self.request is None:
            raise BrowserStateError("no URL has yet been .open()ed")
        return self.open(self.request)

    def back(self):
        """Go back one step in history, and return response object."""
        try:
            self.request, self.response = self._history.pop()
        except IndexError:
            raise BrowserStateError("already at start of history")
        self._parse_html(self.response)
        return self.response

    def links(self, *args, **kwds):
        """Return iteratable over links (mechanize.Link objects)."""
        if not self.viewing_html():
            raise BrowserStateError("not viewing HTML")
        if args:
            raise ValueError("keyword arguments only, please!")
        if kwds:
            return self._find_links(False, **kwds)
        return self._links

    def forms(self):
        """Return iteratable over forms.

        The returned form objects implement the ClientForm.HTMLForm interface.

        """
        if not self.viewing_html():
            raise BrowserStateError("not viewing HTML")
        return self._forms

    def viewing_html(self):
        """Return whether the current response contains HTML data."""
        if self.response is None:
            raise BrowserStateError("not viewing any document")
        return self.response.info().get("content-type") == "text/html"

    def title(self):
        """Return string , with.

        Tags are stripped or textified as described in docs for
        PullParser.get_text() method of pullparser module.

        """
        if not self.viewing_html():
            raise BrowserStateError("not viewing HTML")
        return self._title

    def select_form(self, name=None, predicate=None, nr=None):
        """Select an HTML form for input.

        This is like giving a form the "input focus" in a browser.

        If a form is selected, the object supports the HTMLForm interface, so
        you can call methods like .set_value(), .set(), and .click().

        At least one of the name, predicate and nr arguments must be supplied.
        If no matching form is found, mechanize.FormNotFoundError is raised.

        If name is specified, then the form must have the indicated name.

        If predicate is specified, then the form must match that function.  The
        predicate function is passed the HTMLForm as its single argument, and
        should return a boolean value indicating whether the form matched.

        nr, if supplied, is the sequence number of the form (where 0 is the
        first).  Note that control 0 is the first form matching all the other
        arguments (if supplied); it is not necessarily the first control in the
        form.

        """
        if not self.viewing_html():
            raise BrowserStateError("not viewing HTML")
        if (name is None) and (predicate is None) and (nr is None):
            raise ValueError(
                "at least one argument must be supplied to specify form")

        orig_nr = nr
        for form in self._forms:
            if name is not None and name != form.name:
                continue
            if predicate is not None and not predicate(form):
                continue
            if nr:
                nr -= 1
                continue
            self.form = form
            break  # success
        else:
            # failure
            description = []
            if name is not None: description.append("name '%s'" % name)
            if predicate is not None:
                description.append("predicate %s" % predicate)
            if orig_nr: description.append("nr %d" % orig_nr)
            description = ", ".join(description)
            raise FormNotFoundError("no form matching "+description)

    def _add_referer_header(self, request):
        if (not self._handle_referer or
            self.request.get_type() not in ["http", "https"]):
            return
        try:
            previous_request, previous_response = self._history[-1]
        except IndexError:
            pass
        else:
            previous_url = previous_request.get_full_url()
            request.add_unredirected_header("Referer", previous_url)

    def click(self, *args, **kwds):
        """See ClientForm.HTMLForm.click for documentation."""
        if not self.viewing_html():
            raise BrowserStateError("not viewing HTML")
        request = self.form.click(*args, **kwds)
        self._add_referer_header(request)
        return request

    def submit(self, *args, **kwds):
        """Submit current form.

        Arguments are as for ClientForm.HTMLForm.click().

        """
        return self.open(self.click(*args, **kwds))

    def click_link(self, **kwds):
        """Find a link and return a Request object for it.

        Arguments are as for .find_link().

        """
        if not self.viewing_html():
            raise BrowserStateError("not viewing HTML")
        request = ClientCookie.Request(self.find_link(**kwds).absolute_url)
        self._add_referer_header(request)
        return request

    def follow_link(self, **kwds):
        """Find a link and .open() it.

        Arguments are as for .find_link().

        """
        return self.open(self.click_link(**kwds))

    def find_link(self, *args, **kwds):
        """Find a link in current page.

        Links are returned as mechanize.Link objects.

        # Return third link that .search()-matches the regexp "python"
        # (by ".search()-matches", I mean that the regular expression method
        # .search() is used, rather than .match()).
        find_link(text_regex=re.compile("python"), nr=2)

        # Return first http link in the current page that points to somewhere
        # on python.org whose link text (after tags have been removed) is
        # exactly "monty python" have been removed).
        find_link(text="monty python",
                  url_regex=re.compile("http.*python.org"))

        # Return first link with exactly three HTML attributes.
        find_link(predicate=lambda link: len(link.attrs) == 3)

        Links include anchors (<a>), image maps (<area>), and frames (<frame>,
        <iframe>).

        All arguments must be passed by keyword, not position.  Zero or more
        arguments may be supplied.  In order to find a link, all arguments
        supplied must match.

        If a matching link is not found, mechanize.LinkNotFoundError is raised.

        text: link text between link tags: eg. <a href="blah">this bit</a> (as
         returned by pullparser.get_compressed_text(), ie. without tags but
         with opening tags "textified" as per the pullparser docs) must compare
         equal to this argument, if supplied
        text_regex: link text between tag (as defined above) must match the
         regular expression object passed as this argument, if supplied
        name, name_regex: as for text and text_regex, but matched against the
         name HTML attribute of the link tag
        url, url_regex: as for text and text_regex, but matched against the
         URL of the link tag (note this matches against Link.url, which is a
         relative or absolute URL according to how it was written in the HTML)
        predicate: a function taking a Link object as its single argument,
         returning a boolean result, indicating whether the links
        nr: matches the nth link that matches all other criteria (default 0)

        """
        if args:
            raise ValueError("keyword arguments only, please!")
        return self._find_links(True, **kwds)

    def __getattr__(self, name):
        # pass through ClientForm / DOMForm methods and attributes
        if self.form is not None:
            try: return getattr(self.form, name)
            except AttributeError: pass
        raise AttributeError("%s instance has no attribute %s "
                             "(perhaps you forgot to .select_form()?" %
                             (self.__class__, name))

#---------------------------------------------------
# Private methods.

    def _find_links(self, single,
                    text=None, text_regex=None,
                    name=None, name_regex=None,
                    url=None, url_regex=None,
                    tag=None,
                    predicate=None,
                    nr=0
                    ):
        if not self.viewing_html():
            raise BrowserStateError("not viewing HTML")

        links = []
        orig_nr = nr

        for link in self._links:
            if url is not None and url != link.url:
                continue
            if url_regex is not None and not url_regex.search(link.url):
                continue
            if (text is not None and
                (link.text is None or text != link.text)):
                continue
            if (text_regex is not None and
                (link.text is None or not text_regex.search(link.text))):
                continue
            if name is not None and name != dict(link.attrs).get("name"):
                continue
            if name_regex is not None:
                link_name = dict(link.attrs).get("name")
                if link_name is None or not name_regex.search(link_name):
                    continue
            if tag is not None and tag != link.tag:
                continue
            if predicate is not None and not predicate(link):
                continue
            if nr:
                nr -= 1
                continue
            if single:
                return link
            else:
                links.append(link)
                nr = orig_nr
        if not links:
            raise LinkNotFoundError()
        return links

    def _parse_html(self, response):
        if not self.viewing_html():
            # nothing to see here
            return

        # set ._forms, ._title, ._links
        self._forms = ClientForm.ParseResponse(response)
        self.response.seek(0)

        p = pullparser.PullParser(response)
        try:
            p.get_tag("title")
            self._title = p.get_text()
        except pullparser.NoMoreTokensError:
            self._title = None

        base = self.response.geturl()

        self._links = []
        for token in p.tags(*self.urltags.keys()):
            if token.type == "endtag":
                continue
            attrs = dict(token.attrs)
            tag = token.data
            name = attrs.get("name")
            text = None
            url = attrs.get(self.urltags[tag])
            if tag == "a":
                if token.type != "startendtag":
                    # XXX hmm, this'd break if end tag is missing
                    text = p.get_compressed_text(("endtag", tag))
                # but this doesn't work for eg. <a href="blah"><b>Andy</b></a>
                #text = p.get_compressed_text()
                # This is a hack from WWW::Mechanize to get some really basic
                # JavaScript working, which I'm not yet convinced is a good
                # idea.
##                 onClick = attrs["onclick"]
##                 m = re.search(r"/^window\.open\(\s*'([^']+)'/", onClick)
##                 if onClick and m:
##                     url = m.group(1)
            if not url:
                # Probably an <A NAME="blah"> link or <AREA NOHREF...>.
                # For our purposes a link is something with a URL, so ignore
                # this.
                continue

            link = Link(base, url, text, tag, token.attrs)
            self._links.append(link)

        self.response.seek(0)
